(ns kube-deploy.helpers
  (:require [lambdakube.core :as lk]
            [clojure.data.codec.base64 :as b64]))

(defn kube-namespace
  ([name]
   {:apiVersion "v1"
    :kind "Namespace"
    :metadata {:name name}}))

(defn secrets
  ([name labels options]
   {:apiVersion "v1"
    :kind "Secret"
    :type "Opaque"
    :metadata {:name name :labels labels}
    :data options}))

(defn secrets-registry
  ([name labels options]
   {:apiVersion "v1"
    :kind "Secret"
    :type "kubernetes.io/dockerconfigjson"
    :metadata {:name name :labels labels}
    :data options}))

(defn ingress
  ([name labels]
   (ingress name labels {}))
  ([name labels spec]
   {:apiVersion "networking.k8s.io/v1beta1"
    :kind "Ingress"
    :metadata {:name name :labels labels}
    :spec spec}))

(defn persistent-volume
  ([obj name labels]
   (ingress name labels {}))
  ([obj name labels spec]
   (assoc obj {:apiVersion "v1"
               :kind "PersistentVolume"
               :metadata {:name name :labels labels}
               :spec spec})))


(defn add-to-spec
  ([container name value]
   (-> container
       (assoc name value))))

(defn add-readiness-probe-get
  ([container url port]
   (add-readiness-probe-get container url port 60 60))
  ([container url port delay period]
   (-> container
       (assoc :readinessProbe
              {:periodSeconds period
               :initialDelaySeconds delay
               :failureThreshold 5
               :httpGet {:path url :port port}}))))

(defn add-readiness-probe-cmd
  ([container command]
   (add-readiness-probe-cmd container command 60 60))
  ([container command delay period]
   (-> container
       (assoc :readinessProbe
              {:periodSeconds period
               :initialDelaySeconds delay
               :failureThreshold 5
               :exec {:command command}}))))

(defn add-liveness-probe-get
  ([container url port]
   (add-liveness-probe-get container url port 60 60))
  ([container url port delay period]
   (-> container
       (assoc :livenessProbe
              {:periodSeconds period
               :initialDelaySeconds delay
               :failureThreshold 5
               :httpGet {:path url :port port}}))))

(defn add-liveness-probe-cmd
  ([container command]
   (add-liveness-probe-cmd container command 60 60))
  ([container command delay period]
   (-> container
       (assoc :livenessProbe
              {:periodSeconds delay
               :initialDelaySeconds period
               :failureThreshold 5
               :exec {:command command}}))))

(defn add-tls [obj key val]
  (-> obj
      (update-in [:spec :tls] assoc key val)))

(defn add-rules [obj key val]
  (-> obj
      (update-in [:spec :rules] assoc key val)))

    ;; livenessProbe:
    ;;   exec:
    ;;     command:
    ;;     - cat
    ;;     - /tmp/healthy
    ;;   initialDelaySeconds: 5
    ;;   periodSeconds: 5

;; (defn add-spec [obj key val]
;;   (-> obj
;;       (update-in [:spec] assoc key val)))


(defn add-spec [obj key val]
  (-> obj
      (update-in [:spec] assoc key val)))

(defn decode-secret [value]
  (String. (b64/decode (.getBytes value))))

(defn encode-secret [value]
  (b64/encode (.getBytes value)))

(defn encode-secret-map [data]
  (reduce-kv
   (fn [m k v] (assoc m k (encode-secret v)))
   {}
   data))

(defn to-yaml-store [$ filename]
  (let [y (lk/to-yaml $)]
    (spit filename y) y))

(defn to-yaml [$]
  (let [y (lk/to-yaml $)]
    (spit "test.yaml" y)
    y))


;; (defn add-init-container
;;   ([pod name image options]
;;    (let [container (-> options
;;                        (merge {:name name
;;                                :image image}))]
;;      (update pod :spec field-conj :initContainers container)))
;;   ([pod name image]
;;    (add-init-container pod name image {})))

;; (defn update-container [pod cont-name f & args]
;;   (let [update-cont (fn [cont]
;;                       (if (= (:name cont) cont-name)
;;                         (apply f cont args)
;;                         ;; else
;;                         cont))]
;;     (-> pod
;;         (update-in [:spec :containers] #(map update-cont %))
;;         (update-in [:spec :initContainers] #(map update-cont %))
;;         (update :spec #(if (empty? (:initContainers %))
;;                          (dissoc % :initContainers)
;;                          ;; else
;;                          %)))))


;; (defn wait-for-service-port
;;   ([pod dep portname]
;;    (let [{:keys [hostname ports]} dep
;;          cont (keyword (str "wait-for-" (name hostname) "-" (name portname)))]
;;      (-> pod
;;          (lk/add-init-container cont "busybox"
;;                                 {:command ["sh"
;;                                            "-c"
;;                                            (str "while ! nc -z " hostname " " (ports portname) "; do sleep 1; done")]}))))
;;   ([pod dep]
;;    (when-not (= (-> dep :ports count) 1)
;;      (throw (Exception. "Port name must be specified when waiting for a service exposing more than one port.")))
;;    (let [portname (-> dep :ports first first)]
;;      (wait-for-service-port pod dep portname))))




